/*
	Author: wamesbury
	Created: 2020-04-03
	Purpose: Documentation Requirement Set editor & sub-editors
	History:
	2021-11-17 WRA ND-11641 Added "NOA Required?""
	2021-10-26 WRA ND-11583 Removed forced OasisRequired for PPS billing types. Removed forced HippsRequired for OasisRequired = "Yes"
	2021-03-17 WRA ND-10488 Removed Claim Mapping
	2020-12-08 RMW ND-10622 Added new fields
	2020-06-12 WRA ND-10204 Added restrictions on certification/billing period options
	2020-05-26 WRA ND-10153 Disabling "HIPPS Required" when loading PPS Billing Type set.
*/

define([
	"dojo/_base/declare",
	"thornberry/object/ObjectEditor",
	"thornberry/object/ObjectModel",
	"thornberry/serverRequest",
	"thornberry/object/TableManager",
	"dojo/_base/array",
	"dojo/_base/lang",
	"dojo/aspect",
	"dojo/store/Memory",
	"dojo/dom-construct",
	"dojo/dom-style",
	"dojo/query",
	"dojo/on",
	"dojo/Deferred",
	"dijit/form/CurrencyTextBox",
	"dijit/form/NumberTextBox",
	"dijit/form/DropDownButton",
	"dojox/form/CheckedMultiSelect",
	"dijit/form/MultiSelect",
	"dijit/form/Button",
	"dijit/TooltipDialog",
	"thornberry/object/template",
	"thornberry/EditorCollection"
], function(declare, ObjectEditor, ObjectModel, serverRequest, TableManager, arrayUtils, lang, aspect, Memory,
			domConstruct, domStyle, query, on, Deferred, CurrencyTextBox, NumberTextBox, DropDownButton,
			CheckedMultiSelect, MultiSelect, Button, TooltipDialog, template, EditorCollection) {

	var insuranceTypeOptions = {};
	TableManager.get("FillInGroup~20605").query().forEach(function(type) {
		insuranceTypeOptions[type.id] = type.name;
	});
	if (hospiceAgency != 1) {
		delete insuranceTypeOptions[2];
	}

	var requirementOptions = {};
	TableManager.get("Tables.DocumentationRequirement").query().forEach(function(requirement) {
		requirementOptions[requirement.id] = requirement.name;
	});

	TableManager.get("DocumentLibrary.Type").query().forEach(function(type) {
		// Make name a string and "Id" lowercase
		type.Name = type.Name.toString();
		type.id = type.Id;
	});

	TableManager.add("DayOfWeek", new Memory({ data: [
		{ "id": "3", "name": "Sunday" },
		{ "id": "4", "name": "Monday" },
		{ "id": "5", "name": "Tuesday" },
		{ "id": "6", "name": "Wednesday" },
		{ "id": "0", "name": "Thursday" },
		{ "id": "1", "name": "Friday" },
		{ "id": "2", "name": "Saturday" }]
	}));

	var docRequirementSetModel = new ObjectModel({ propertyDefinitions: [
		{ name: "id", type: "text" },
		{ name: "Name", type: "text", maxLength: 50, required: true },
		{ name: "Status", type: "enum", displayIndex: { "active": "Active", "inactive": "Inactive" }, required: true },
		{ name: "InsuranceType", type: "enum", displayIndex: insuranceTypeOptions, required: true },
		{ name: "BillingType", type: "tableItem", tableName: "FillInGroup~20606", required: true },
		{ name: "CertPeriodLengthType", type: "tableItem", tableName: "FillInGroup~20607", required: true },
		{ name: "CertPeriodLengthFirst", type: "text" },
		{ name: "CertPeriodLengthSecond", type: "text" },
		{ name: "CertPeriodLengthThird", type: "text" },
		{ name: "CertPeriodLengthSubsequent", type: "text" },
		{ name: "BillPeriodLengthType", type: "tableItem", tableName: "FillInGroup~20608", required: true },
		{ name: "BillPeriodAmount", type: "text" },
		{ name: "BillPeriodFactor", type: "tableItem", tableName: "FillInGroup~20609" },
		{ name: "BillPeriodWeekEnd", type: "tableItem", tableName: "DayOfWeek" },
		{ name: "BillPeriodSpanCalendarMonth", type: "boolean" },
		{ name: "OasisRequired", type: "boolean" },
		{ name: "HippsRequired", type: "boolean" },
		{ name: "GrouperType", type: "tableItem", tableName: "FillInGroup~20610" },
		{ name: "PricerType", type: "tableItem", tableName: "FillInGroup~20611" },
		{ name: "PricerFlatRate", type: "text" },
		{ name: "PocRequired", type: "boolean" },
		{ name: "F2FRequired", type: "boolean" },
		{ name: "NoeRequired", type: "boolean" },
		{ name: "NotrRequired", type: "boolean" },
		{ name: "NoaRequired", type: "boolean" },
		{ name: "RapRequired", type: "boolean" },
		{ name: "RapInitialPaymentPercentage", type: "text" },
		{ name: "RapSubsequentPaymentPercentage", type: "text" },
		{ name: "RapTimelyFilingTimeAllowed", type: "text" },
		{ name: "RapTimelyFilingTimeAllowedUnit", type: "tableItem", tableName: "FillInGroup~20627" },
		{ name: "RapTimelyFilingFollowUp", type: "text" },
		{ name: "RapTimelyFilingFollowUpUnit", type: "tableItem", tableName: "FillInGroup~20628" },
		{ name: "RapOasisRequired", type: "boolean" },
		{ name: "RapPocRequired", type: "boolean" },
		{ name: "RapAllowSecondPeriodRap", type: "boolean" }
	]});

	var docRequirementSetTemplate = {
		noForm: false,
		containers: [{
			sections: [{
				hideDivider: false,
				rows: [{
					fields: [{
						name: "Insurances",
						displayName: "Insurances:",
						labelSize: "medium-large"
					}, {
						name: "copyButton",
						containerStyle: "vertical-align: top;",
						noField: true
					}]
				}]
			}, {
				hideDivider: false,
				rows: [{
					fields: [{
						name: "Name",
						displayName: "Name:",
						fieldSize: "largest",
						labelSize: "medium-large"
					}, {
						name: "Status",
						displayName: "Status:",
						labelSize: "smaller"
					}]
				}, {
					fields: [{
						name: "InsuranceType",
						displayName: "Insurance Type:",
						labelSize: "medium-large"
					}]
				}, {
					fields: [{
						name: "BillingType",
						displayName: "Billing Type:",
						labelSize: "medium-large"
					}]
				}]
			}, {
				title: "Certification Periods",
				rows: [{
					fields: [{
						name: "CertPeriodLengthType",
						displayName: "Period Length:",
						fieldSize: "large"
					}, {
						name: "CertPeriodLengthFirst",
						displayName: "First:",
						labelSize: "medium",
						fieldSize: "small"
					}]
				}, {
					fields: [{
						name: "spacer1",
						noField: true,
						fieldSize: "large"
					}, {
						name: "CertPeriodLengthSecond",
						displayName: "Second:",
						labelSize: "medium",
						fieldSize: "small"
					}]
				}, {
					fields: [{
						name: "spacer2",
						noField: true,
						fieldSize: "large"
					},{
						name: "CertPeriodLengthThird",
						displayName: "Third:",
						labelSize: "medium",
						fieldSize: "small"
					}]
				}, {
					fields: [{
						name: "spacer3",
						noField: true,
						fieldSize: "large"
					},{
						name: "CertPeriodLengthSubsequent",
						displayName: "Subsequent:",
						labelSize: "medium",
						fieldSize: "small"
					}]
				}]
			}, {
				title: "Billing Periods",
				rows: [{
					fields: [{
						name: "BillPeriodLengthType",
						displayName: "Bill Period Length:",
						labelSize: "medium-larger",
						fieldSize: "medium-small"
					}, {
						name: "BillPeriodAmount",
						displayName: "Custom Length:",
						fieldSize: "smaller"
					}, {
						name: "BillPeriodFactor",
						hideLabel: true
					}]
				}, {
					fields: [{
						name: "BillPeriodWeekEnd",
						displayName: "Week ends on:",
						labelSize: "medium-larger"
					}]
				}, {
					fields: [{
						name: "BillPeriodSpanCalendarMonth",
						displayName: "Period can span calendar months?",
						labelSize: "largest"
					}]
				}]
			}, {
				title: "General",
				rows: [{
					fields: [{
						name: "OasisRequired",
						displayName: "OASIS Required?",
						labelSize: "medium-larger",
						fieldSize: "small"
					}]
				}, {
					fields: [{
						name: "HippsRequired",
						displayName: "HIPPS Required?",
						labelSize: "medium-larger",
						fieldSize: "small"
					}, {
						name: "GrouperType",
						displayName: "Grouper:",
						labelSize: "small",
						fieldSize: "medium-large"
					}, {
						name: "PricerType",
						displayName: "Pricer:",
						labelSize: "smaller"
					}, {
						name: "PricerFlatRate",
						hideLabel: true,
						fieldSize: "small"
					}]
				}, {
					fields: [{
						name: "PocRequired",
						displayName: "POC Required?",
						labelSize: "medium-larger",
						fieldSize: "small"
					}]
				}, {
					fields: [{
						name: "F2FRequired",
						displayName: "F2F Required?",
						labelSize: "medium-larger",
						fieldSize: "small"
					}]
				}, {
					fields: [{
						name: "NoeRequired",
						displayName: "NOE Required?",
						labelSize: "medium-larger",
						fieldSize: "small"
					}]
				}, {
					fields: [{
						name: "NotrRequired",
						displayName: "NOTR Required?",
						labelSize: "medium-larger",
						fieldSize: "small"
					}]
				}, {
					fields: [{
						name: "NoaRequired",
						displayName: "NOA Required?",
						labelSize: "medium-larger",
						fieldSize: "small"
					}]
				}]
			}]
		}, {
			attachPoint: "ppsRapContainer",
			sections: [{
				title: "PPS RAP Requirements",
				rows: [{
					fields: [{
						name: "RapRequired",
						displayName: "RAP Required?",
						labelSize: "larger",
						fieldSize: "small"
					}]
				}, {
					fields: [{
						name: "RapInitialPaymentPercentage",
						displayName: "Initial RAP %",
						labelSize: "larger",
						fieldSize: "small"
					}, {
						name: "RapSubsequentPaymentPercentage",
						displayName: "Subsequent RAP %",
						labelSize: "large",
						fieldSize: "small"
					}]
				}, {
					fields: [{
						name: "RapTimelyFilingTimeAllowed",
						displayName: "Timely Filing Time Allowed",
						labelSize: "larger",
						fieldSize: "small"
					}, {
						name: "RapTimelyFilingTimeAllowedUnit",
						displayName: "",
						hideLabel: true,
						fieldSize: "small"
					}]
				}, {
					fields: [{
						name: "RapTimelyFilingFollowUp",
						displayName: "Default Follow Up Date",
						labelSize: "larger",
						fieldSize: "small"
					}, {
						name: "RapTimelyFilingFollowUpUnit",
						displayName: "",
						hideLabel: true,
						fieldSize: "small"
					}]
				}, { 
					fields: [{
						name: "RapOasisRequired",
						displayName: "OASIS Required for RAP?",
						labelSize: "larger",
						fieldSize: "small"
					}]
				}, {
					fields: [{
						name: "RapPocRequired",
						displayName: "POC Required for RAP?",
						labelSize: "larger",
						fieldSize: "small"
					}]
				}, {
					fields: [{
						name: "RapAllowSecondPeriodRap",
						displayName: "Allow Period 2 RAP during Period 1?",
						labelSize: "larger",
						fieldSize: "small"
					}]
				}, {
					containerAttachPoint: "ppsRapDocumentRequirements"
				}, {
					containerAttachPoint: "ppsRapRequirements"
				}]
			}]
		}, {
			attachPoint: "billingTypeRequirementsContainer",
			sections: [{
				title: "Claim Requirements",
				rows: [{
					containerAttachPoint: "billingTypeDocumentRequirements"
				}, {
					containerAttachPoint: "billingTypeRequirements"
				}]
			}]
		}]
	};

	DocRequirementSet = docRequirementSetModel.getObject();
	var DocRequirementSetEditor = declare([ObjectEditor], {
		objectModel: docRequirementSetModel,
		templateString: template.generate(docRequirementSetTemplate),
		requirementCollections: [],
		documentReqCollections: [],

		postMixInProperties: function() {
			var t = this;
			t.inherited("postMixInProperties", arguments);

			var toCurrencyTextBox = function(fieldName) {
				var propertyModel = t.getPropertyModel(fieldName);
				propertyModel.getField = function(node) {
					return new CurrencyTextBox(lang.mixin(this.getFieldProperties(), { currency: "USD" }), node);
				};
				propertyModel.getInternal = function(value) {
					return (value ? value : (value === 0 ? value : ""));
				};
			};

			var toNumberTextBox = function(fieldName) {
				var propertyModel = t.getPropertyModel(fieldName);
				propertyModel.getField = function(node) {
					return new NumberTextBox(this.getFieldProperties(), node);
				};
				propertyModel.getInternal = function(value) {
					return (value ? value : (value === 0 ? value : ""));
				};
			};

			arrayUtils.map(["CertPeriodLengthFirst", "CertPeriodLengthSecond", "CertPeriodLengthThird",
							"CertPeriodLengthSubsequent", "BillPeriodAmount", "RapInitialPaymentPercentage",
							"RapSubsequentPaymentPercentage"], toNumberTextBox);
			toCurrencyTextBox("PricerFlatRate");

			var certLengthProps = { invalidMessage: "30-180 days",
									constraints: {fractional: false, min: 30, max: 180} };

			lang.mixin(t.getPropertyModel("CertPeriodLengthFirst").fieldProperties, certLengthProps);
			lang.mixin(t.getPropertyModel("CertPeriodLengthSecond").fieldProperties, certLengthProps);
			lang.mixin(t.getPropertyModel("CertPeriodLengthThird").fieldProperties, certLengthProps);
			lang.mixin(t.getPropertyModel("CertPeriodLengthSubsequent").fieldProperties, certLengthProps);

			t.getPropertyModel("BillPeriodAmount").fieldProperties.constraints = {fractional: false, min: 1, max: 366};
			t.getPropertyModel("RapInitialPaymentPercentage").fieldProperties.constraints = {fractional: false, min: 0, max: 100};
			t.getPropertyModel("RapSubsequentPaymentPercentage").fieldProperties.constraints = {fractional: false, min: 0, max: 100};

			t.getPropertyModel("InsuranceType").noDefault = true;
		},

		buildRendering: function() {
			var t = this;
			t.inherited("buildRendering", arguments);

			t.Insurances = new MultiSelect({style: "min-width: 300px;"}, t.InsurancesFieldContainer);

			t.copyDocRequirementSetButton = new Button({
				label: "Copy Documentation Requirement Set",
				onClick: function() {
					var saveData = t._getSaveData();
					t.pageEditor.object.set("editMode", "add");
					delete saveData.Name;
					delete saveData.id;
					t.set("object", new DocRequirementSet(saveData));
				}
			},t.copyButtonLabelContainer);

			domStyle.set(t.CertPeriodLengthFirstField.domNode, "width", "50px");
			domStyle.set(t.CertPeriodLengthSecondField.domNode, "width", "50px");
			domStyle.set(t.CertPeriodLengthThirdField.domNode, "width", "50px");
			domStyle.set(t.CertPeriodLengthSubsequentField.domNode, "width", "50px");
			domStyle.set(t.BillPeriodAmountField.domNode, "width", "50px");
			domStyle.set(t.RapInitialPaymentPercentageField.domNode, "width", "50px");
			domStyle.set(t.RapSubsequentPaymentPercentageField.domNode, "width", "50px");
			domStyle.set(t.PricerFlatRateField.domNode, "width", "80px");
			domStyle.set(t.RapTimelyFilingTimeAllowedField.domNode, "width", "50px");
			domStyle.set(t.RapTimelyFilingFollowUpField.domNode, "width", "50px");
		},

		postCreate: function() {
			var t = this;
			t.inherited(arguments);

			domConstruct.create("label", { innerHTML: "&nbsp;days" }, t.CertPeriodLengthFirstFieldContainer);
			domConstruct.create("label", { innerHTML: "&nbsp;days" }, t.CertPeriodLengthSecondFieldContainer);
			domConstruct.create("label", { innerHTML: "&nbsp;days" }, t.CertPeriodLengthThirdFieldContainer);
			domConstruct.create("label", { innerHTML: "&nbsp;days" }, t.CertPeriodLengthSubsequentFieldContainer);

			var populateBillPeriodLengthType = function() {
				var billingType = t.object.get("BillingType");
				var certType = t.object.get("CertPeriodLengthType");

				var restrictIdArray = [];
				TableManager.get("FillInGroup~20608").query().forEach(function(lengthType) {
					// When Cert Period = No Recertification Required, Billing Periods should not be allowed to be Cert Period
					// Include PDGM bill period length only when billing type is PPS
					// When Billing Type = PPS, Billing Periods can only be Cert Period or PDGM
					if (((certType != 1) || (lengthType.id != 1)) && 
						(((billingType == 1) && ((lengthType.id == 1) || (lengthType.id == 2))) ||
						((billingType != 1) && ((lengthType.id != 2))))) {
						restrictIdArray.push(lengthType.id);
					}
				});

				// Remove current bill period length selection if invalid
				var newBillPeriodLengthTypeVal = t.object.get("BillPeriodLengthType");
				if (arrayUtils.indexOf(restrictIdArray, newBillPeriodLengthTypeVal) == -1) {
					newBillPeriodLengthTypeVal = "";
				}
				t.populateSelect(t.BillPeriodLengthTypeField, TableManager.get("FillInGroup~20608"),
								 newBillPeriodLengthTypeVal, true, restrictIdArray);
				t.object.set("BillPeriodLengthType", newBillPeriodLengthTypeVal);
			};

			// Enable Pricer if a grouper is selected
			// When Billing Type = FFS, disable/hide Pricer
			var enablePricerLogic = function() {
				var billingType = t.object.get("BillingType");
				var grouperType = t.object.get("GrouperType");
				
				if ((grouperType == "") || (billingType == 3)) {
					t.disable("PricerType", true);
					t.object.set("PricerType", "");
				} else {
					t.disable("PricerType", false);
				}
			};

			t.own(aspect.before(t, "_setObjectAttr", function(object) {
				var editorCollections = t.requirementCollections.concat(t.documentReqCollections);
				arrayUtils.forEach(editorCollections, function(collection) {
					collection.destroyRecursive();
				});
				
				t.requirementCollections = [];
				t.documentReqCollections = [];

				if (!object.InsuranceType) {
					object.InsuranceType = 1; // default to Homecare
				}
			}));

			t.own(aspect.after(t, "_setObjectAttr", function(object) {
				// Populate Insurances with those that use this documentation requirement set
				query("option", t.Insurances.containerNode).forEach(function(option) {domConstruct.destroy(option);});
				TableManager.get("Tables.Insurance.DocumentationRequirement.Assignment").query({
					requirementSet: object.id
				}).forEach(function(assignment) {
					domConstruct.create("option", {
						label:escapeText(assignment.name), id:assignment.id, disabled: true
					}, t.Insurances.containerNode);
				});
			}, true));

			t.registerChangeLogic("InsuranceType", function(value) {
				// If (Billing Type? PPS), set HippsRequired to "Yes"
				if (!t.inObjectLoad) {
					if (value != 1) {
						t.object.set("HippsRequired", "");
					} else if (t.object.get("BillingType") == 1) {
						t.object.set("HippsRequired", 1);
					}
				}

				var restrictIdArray = [3];	 // All get FFS
				if (value == 1) {
					restrictIdArray.push(1); // Homecare gets PPS
				} else if (value == 2) {
					restrictIdArray.push(2); // Hospice gets Per Diem
				}

				// Remove current bill period length selection if invalid
				var newBillPeriodTypeVal = t.object.get("BillingType");
				if (arrayUtils.indexOf(restrictIdArray, newBillPeriodTypeVal) == -1) {
					newBillPeriodTypeVal = "";
				}
				t.populateSelect(t.BillingTypeField, TableManager.get("FillInGroup~20606"),
								 newBillPeriodTypeVal, true, restrictIdArray);
				t.object.set("BillingType", newBillPeriodTypeVal);

				var displayHomecare = (value == 1) ? "" : "none";
				var displayHospice = (value == 2) ? "" : "none";
				domStyle.set(t.OasisRequiredContainer, "display", displayHomecare);
				domStyle.set(t.HippsRequiredContainer, "display", displayHomecare);
				domStyle.set(t.GrouperTypeContainer, "display", displayHomecare);
				domStyle.set(t.PricerTypeContainer, "display", displayHomecare);
				domStyle.set(t.PricerFlatRateContainer, "display", displayHomecare);
				domStyle.set(t.NotrRequiredContainer, "display", displayHospice);
				domStyle.set(t.NoeRequiredContainer, "display", displayHospice);
				domStyle.set(t.NoaRequiredContainer, "display", displayHomecare);
				
				if (value == 1) {
					t.object.set("NoeRequired", "");
					t.object.set("NotrRequired", "");
				} else if (value == 2) {
					t.object.set("NoaRequired", "");
				}
				t.disable("NoeRequired", (value == 1));
				t.disable("NotrRequired", (value == 1));
				t.disable("NoaRequired", (value == 2));
			});

			t.registerChangeLogic("BillingType", function(value) {
				populateBillPeriodLengthType();
				enablePricerLogic();

				// When Billing Type = PPS, Cert Periods should be forced to be 60-days
				if (value == 1) {
					t.object.set("CertPeriodLengthType", 2);
				}
				t.disable("CertPeriodLengthType", (value == 1));

				// Show PPS RAP Requirements when Billing Type = PPS
				domStyle.set(t.ppsRapContainer, "display", value == 1 ? "" : "none");
				if (value != 1) {
					t.object.set("RapRequired", 0);
					if (t.ppsRapRequirementEditors) {
						t.ppsRapRequirementEditors.removeAll();
					}
				}

				//  FFS/PPS Final/Per Diem claim requirements
				t.requirementCollections = [];
				t.documentReqCollections = [];
				var containerTitle;
				if (value == 1) {
					t.addRequirements("billingType", "ppsFinal", t.billingTypeRequirements);
					t.addDocumentRequirements("billingType", t.billingTypeDocumentRequirements);
					containerTitle = "PPS Final Requirements";
				} if (value == 2) {
					t.addRequirements("billingType", "hospice", t.billingTypeRequirements);
					t.addDocumentRequirements("billingType", t.billingTypeDocumentRequirements);
					containerTitle = "Per Diem Requirements";
				} else if (value == 3) {
					t.addRequirements("billingType", "ffs", t.billingTypeRequirements);
					t.addDocumentRequirements("billingType", t.billingTypeDocumentRequirements);
					containerTitle = "FFS Requirements";
				}
				domStyle.set(t.billingTypeRequirementsContainer, "display", +value ? "" : "none");
				var titleNode = query(".content-area-title-text", t.billingTypeRequirementsContainer)[0];
				if (titleNode) {
					titleNode.innerHTML = containerTitle;
				}
				
				// OasisRequired/HippsRequired logic
				if (!t.inObjectLoad) {
					if (value == 1) { // PPS
						// if (Billing Type? PPS), set OasisRequired to "Yes"
						t.object.set("OasisRequired", 1);

						if (t.object.get("InsuranceType") == 1) {
							t.object.set("HippsRequired", 1);
						}
					} else if (value == 3) { // FFS
						// If (Billing Type? FFS), set OasisRequired to ""
						t.object.set("OasisRequired", 0);

						if (t.object.get("InsuranceType") == 1) {
							t.object.set("HippsRequired", 0);
						}
					}
				}
			});

			t.registerChangeLogic("CertPeriodLengthType", function(value) {
				populateBillPeriodLengthType();
				
				// "Custom"
				t.disable("CertPeriodLengthFirst", (value != 4));
				t.disable("CertPeriodLengthSubsequent", (value != 4));

				t.require("CertPeriodLengthFirst", (value == 4));
				t.require("CertPeriodLengthSubsequent", (value == 4));

				if (value != 4) {
					t.object.set("CertPeriodLengthFirst", "");
					t.object.set("CertPeriodLengthSecond", "");
					t.object.set("CertPeriodLengthThird", "");
					t.object.set("CertPeriodLengthSubsequent", "");
				}

				var display = (value == 4) ? "" : "none";
				domStyle.set(t.CertPeriodLengthFirstContainer, "display", display);
				domStyle.set(t.CertPeriodLengthSecondContainer, "display", display);
				domStyle.set(t.CertPeriodLengthThirdContainer, "display", display);
				domStyle.set(t.CertPeriodLengthSubsequentContainer, "display", display);

				domStyle.set(t.spacer1Container, "display", display);
				domStyle.set(t.spacer2Container, "display", display);
				domStyle.set(t.spacer3Container, "display", display);
			});

			t.registerChangeLogic("CertPeriodLengthFirst", function(value) {
				t.disable("CertPeriodLengthSecond", (value == "" || isNaN(value)));
				if (value == "" || isNaN(value)) {
					t.object.set("CertPeriodLengthSecond", "");
				}
			});

			t.registerChangeLogic("CertPeriodLengthSecond", function(value) {
				t.disable("CertPeriodLengthThird", (value == "" || isNaN(value)));
				if (value == "" || isNaN(value)) {
					t.object.set("CertPeriodLengthThird", "");
				}
			});

			t.registerChangeLogic("BillPeriodLengthType", function(value) {
				t.disable("BillPeriodAmount", (value != 4));
				t.disable("BillPeriodFactor", (value != 4));
				t.require("BillPeriodAmount", (value == 4));
				t.require("BillPeriodFactor", (value == 4));

				if (value != 4) {
					t.object.set("BillPeriodAmount", "");
					t.object.set("BillPeriodFactor", "");
				}

				var display = (value == 4) ? "" : "none";
				domStyle.set(t.BillPeriodAmountContainer, "display", display);
				domStyle.set(t.BillPeriodFactorContainer, "display", display);
				domStyle.set(t.BillPeriodWeekEndContainer, "display", display);
				domStyle.set(t.BillPeriodSpanCalendarMonthContainer, "display", display);

				var rapOptionEnabled = (t.object.get("RapRequired") && value == 2);
				domStyle.set(t.RapAllowSecondPeriodRapContainer, "display", (rapOptionEnabled ? "" : "none"));
				t.disable("RapAllowSecondPeriodRap",!rapOptionEnabled);
				if (!rapOptionEnabled) {
					t.object.set("RapAllowSecondPeriodRap", "")
				}
			});

			t.registerChangeLogic("BillPeriodFactor", function(value) {

				t.disable("BillPeriodWeekEnd", (value != 1));
				t.disable("BillPeriodSpanCalendarMonth", (value != 1 && value != 3));

				if (value != 1) {
					t.object.set("BillPeriodWeekEnd", "");
				} else if (t.object.get("BillPeriodWeekEnd") == "") {
					// default to Saturday
					t.object.set("BillPeriodWeekEnd", "2");
				}

				if ((value != 1) && (value != 3)) {
					t.object.set("BillPeriodSpanCalendarMonth", "");
				}
			});

			t.registerChangeLogic("OasisRequired", function(value) {
				// if OASIS is not required, set HippsRequired to "No"
				if (!value && !t.inObjectLoad) {
					t.object.set("HippsRequired", 0);
				}

				// Show/hide requirement option editors that depend on this
				arrayUtils.forEach(t.requirementCollections, function(collection) {
					if (collection !== t["ppsRapRequirementEditors"]) {
						collection.get("editorStore").query().forEach(function(editor) {
							var optionId = editor.object.get("id") || editor.id;
							var option = TableManager.get("Tables.DocumentationRequirementsOption").get(optionId);
							if (option.OasisRequired) {
								domStyle.set(editor.domNode, "display", value ? "" : "none");
							}
						});
					}
				});
			});

			t.registerChangeLogic("HippsRequired", function(value) {
				// Enable Grouper field if HIPPS is required
				if (!value) {
					t.disable("GrouperType", true);
					t.object.set("GrouperType", "");
				} else {
					t.disable("GrouperType", false);
				}
			});

			t.registerChangeLogic("GrouperType", function(value) {
				enablePricerLogic();

				var restrictIdArray = [];
				if (value == 1) {
					// Grouper? Medicare HH-PPS: Medicare HH-PPS
					restrictIdArray = [1,4];
				} else if (value == 2) {
					// Grouper? Medicare HH-PDGM: Medicare HH-PDGM
					restrictIdArray = [2,4];
				} else if (value == 3) {
					// Grouper? NY Medicaid: NY Medicaid
					restrictIdArray = [3,4];
				}
				
				// Remove current selection if invalid
				var newPricerTypeVal = t.object.get("PricerType");
				if (arrayUtils.indexOf(restrictIdArray, newPricerTypeVal) == -1) {
					newPricerTypeVal = "";
				}
				t.populateSelect(t.PricerTypeField, TableManager.get("FillInGroup~20611"),
								 newPricerTypeVal, true, restrictIdArray);
				t.object.set("PricerType", newPricerTypeVal);
			});

			t.registerChangeLogic("PricerType", function(value) {
				// Enable Flat Rate field if pricer type = flat rate
				if (value == 4) {
					t.disable("PricerFlatRate", false);
					t.require("PricerFlatRate", true);
				} else {
					t.disable("PricerFlatRate", true);
					t.require("PricerFlatRate", false);
					t.object.set("PricerFlatRate", "");
				}
			});

			t.registerChangeLogic("PocRequired", function(value) {
				// Show/hide requirement option editors that depend on this
				arrayUtils.forEach(t.requirementCollections, function(collection) {
					if (collection !== t["ppsRapRequirementEditors"]) {
						collection.get("editorStore").query().forEach(function(editor) {
							var optionId = editor.object.get("id") || editor.id;
							var option = TableManager.get("Tables.DocumentationRequirementsOption").get(optionId);
							if (option.PocRequired) {
								domStyle.set(editor.domNode, "display", value ? "" : "none");
							}
						});
					}
				});
			});

			t.registerChangeLogic("F2FRequired", function(value) {
				// Show/hide requirement option editors that depend on this
				arrayUtils.forEach(t.requirementCollections, function(collection) {
					collection.get("editorStore").query().forEach(function(editor) {
						var optionId = editor.object.get("id") || editor.id;
						var option = TableManager.get("Tables.DocumentationRequirementsOption").get(optionId);
						if (option.F2FRequired) {
							domStyle.set(editor.domNode, "display", value ? "" : "none");
						}
					});
				});
			});

			t.registerChangeLogic("NoeRequired", function(value) {
				// Show/hide requirement option editors that depend on this
				arrayUtils.forEach(t.requirementCollections, function(collection) {
					collection.get("editorStore").query().forEach(function(editor) {
						var optionId = editor.object.get("id") || editor.id;
						var option = TableManager.get("Tables.DocumentationRequirementsOption").get(optionId);
						if (option.NoeRequired) {
							domStyle.set(editor.domNode, "display", value ? "" : "none");
						}
					});
				});
			});

			t.registerChangeLogic("NoaRequired", function(value) {
				// Show/hide requirement option editors that depend on this
				arrayUtils.forEach(t.requirementCollections, function(collection) {
					collection.get("editorStore").query().forEach(function(editor) {
						var optionId = editor.object.get("id") || editor.id;
						var option = TableManager.get("Tables.DocumentationRequirementsOption").get(optionId);
						if (option.NoaRequired) {
							domStyle.set(editor.domNode, "display", value ? "" : "none");
						}
					});
				});
			});

			t.registerChangeLogic("RapRequired", function(value) {
				// Show RAP fields if required
				value = +value;

				t.disable("RapInitialPaymentPercentage", !value);
				t.disable("RapSubsequentPaymentPercentage", !value);
				t.disable("RapTimelyFilingTimeAllowed", !value);
				t.disable("RapTimelyFilingFollowUp", !value);
				t.disable("RapPocRequired", !value);
				t.disable("RapOasisRequired", !value);

				t.require("RapInitialPaymentPercentage", value);
				t.require("RapSubsequentPaymentPercentage", value);

				var display = value ? "" : "none";
				domStyle.set(t.RapInitialPaymentPercentageContainer, "display", display);
				domStyle.set(t.RapSubsequentPaymentPercentageContainer, "display", display);
				domStyle.set(t.RapTimelyFilingTimeAllowedContainer, "display", display);
				domStyle.set(t.RapTimelyFilingTimeAllowedUnitContainer, "display", display);
				domStyle.set(t.RapTimelyFilingFollowUpContainer, "display", display);
				domStyle.set(t.RapTimelyFilingFollowUpUnitContainer, "display", display);
				domStyle.set(t.RapPocRequiredContainer, "display", display);
				domStyle.set(t.RapOasisRequiredContainer, "display", display);
				domStyle.set(t.ppsRapDocumentRequirements, "display", display);
				domStyle.set(t.ppsRapRequirements, "display", display);

				// Clear PPS RAP values if not required
				if (value == 0) {
					t.object.set("RapInitialPaymentPercentage", "");
					t.object.set("RapSubsequentPaymentPercentage", "");
					t.object.set("RapTimelyFilingTimeAllowed", "");
					t.object.set("RapTimelyFilingTimeAllowedUnit", "");
					t.object.set("RapTimelyFilingFollowUp", "");
					t.object.set("RapTimelyFilingFollowUpUnit", "");
					t.object.set("RapPocRequired", "");
					t.object.set("RapOasisRequired", "");

					if (t.ppsRapDocumentReqEditors) {
						t.ppsRapDocumentReqEditors.destroyRecursive();
						t.documentReqCollections = arrayUtils.filter(t.documentReqCollections,function(collection) { return collection !== t.ppsRapDocumentReqEditors; });
					}
					if (t.ppsRapRequirementEditors) {
						t.ppsRapRequirementEditors.destroyRecursive();
						t.requirementCollections = arrayUtils.filter(t.requirementCollections,function(collection) { return collection !== t.ppsRapRequirementEditors; });
					}
				} else {
					t.addRequirements("ppsRap", "ppsRap", t.ppsRapRequirements);
					t.addDocumentRequirements("ppsRap", t.ppsRapDocumentRequirements);
				}

				var secondPeriodOptionEnabled = (t.object.get("BillPeriodLengthType") == 2 && value);
				domStyle.set(t.RapAllowSecondPeriodRapContainer, "display", (secondPeriodOptionEnabled ? "" : "none"));
				t.disable("RapAllowSecondPeriodRap",!secondPeriodOptionEnabled);
				if (!secondPeriodOptionEnabled) {
					t.object.set("RapAllowSecondPeriodRap", "")
				}
			});

			t.registerChangeLogic("RapTimelyFilingTimeAllowed", function(value) {
				t.disable("RapTimelyFilingTimeAllowedUnit", !value);
				t.require("RapTimelyFilingTimeAllowedUnit", value);
				if (!value) {
					t.object.set("RapTimelyFilingTimeAllowedUnit", "");
				}
			});

			t.registerChangeLogic("RapTimelyFilingFollowUp", function(value) {
				t.disable("RapTimelyFilingFollowUpUnit", !value);
				t.require("RapTimelyFilingFollowUpUnit", value);
				if (!value) {
					t.object.set("RapTimelyFilingFollowUpUnit", "");
				}
			});

			t.registerChangeLogic("RapOasisRequired", function(value) {
				// Show/hide requirement option editors that depend on this
				arrayUtils.forEach(t.requirementCollections, function(collection) {
					if (collection === t["ppsRapRequirementEditors"]) {
						collection.get("editorStore").query().forEach(function(editor) {
							var optionId = editor.object.get("id") || editor.id;
							var option = TableManager.get("Tables.DocumentationRequirementsOption").get(optionId);
							if (option.OasisRequired) {
								domStyle.set(editor.domNode, "display", value ? "" : "none");
							}
						});
					}
				});
			});

			t.registerChangeLogic("RapPocRequired", function(value) {
				// Show/hide requirement option editors that depend on this
				arrayUtils.forEach(t.requirementCollections, function(collection) {
					if (collection === t["ppsRapRequirementEditors"]) {
						collection.get("editorStore").query().forEach(function(editor) {
							var optionId = editor.object.get("id") || editor.id;
							var option = TableManager.get("Tables.DocumentationRequirementsOption").get(optionId);
							if (option.PocRequired) {
								domStyle.set(editor.domNode, "display", value ? "" : "none");
							}
						});
					}
				});
			});
		},

		addRequirements: function(claimType, requirementType, attachPoint) {
			var t = this;
			var requirementTypeTitleCase = requirementType.substring(0,1).toUpperCase() + requirementType.substring(1);

			var collectionName;
			if (claimType == "ppsRap") {
				collectionName = "ppsRapRequirementEditors";
			} else {
				collectionName = "billingTypeRequirementEditors";
			}
			
			// TODO: reuse existing editors
			if (t[collectionName]) {
				t[collectionName].destroyRecursive();
			}
			t[collectionName] = new EditorCollection({
				Object: Requirement,
				ObjectEditor: RequirementEditor,
				docRequirementSet: t
			});

			t[collectionName]._getValueAttr = function() {
				var t = this;
				var value = [];
				t.get("editorStore").query().forEach(function(editor) {
					// Don't save invalid values
					if (domStyle.get(editor.domNode, "display") != "none") {
						value.push(editor.object.getData());
					}
				});

				return value;
			};

			var query = {};
			query["ClaimType"+requirementTypeTitleCase] = 1;

			TableManager.get("Tables.DocumentationRequirementsOption").query(query).forEach(function(reqOption) {
				var option = lang.clone(reqOption);
				
				// load existing values
				if (t.object.requirementsOptions) {
					arrayUtils.forEach(t.object.requirementsOptions[claimType], function(requirement) {
						if (option.id == requirement.id) {
							lang.mixin(option, requirement);
						}
					});
				}
				
				option.defaultValue = option["Default" + requirementTypeTitleCase] || option.DefaultValue;
				var requirementEditor = t[collectionName].add(option);
				requirementEditor.requirementLabelContainer.innerHTML = option.name;

				if ((option.OasisRequired && (!t.object.get("OasisRequired") || claimType == "ppsRap")) ||
					(option.PocRequired && (!t.object.get("PocRequired") || claimType == "ppsRap")) ||
					(option.F2FRequired && !t.object.get("F2FRequired")) ||
					(option.NoeRequired && !t.object.get("NoeRequired")) ||
					(option.NoaRequired && !t.object.get("NoaRequired"))) {
						domStyle.set(requirementEditor.domNode, "display", "none");
					}
			});
			t[collectionName].startup();
			// Prevent adding new editors
			t[collectionName].actionStore.remove("add");

			domConstruct.place(t[collectionName].domNode, attachPoint, "last");

			t.requirementCollections.push(t[collectionName]);
		},

		addDocumentRequirements: function(claimType, attachPoint) {
			var t = this;

			var collectionName;
			if (claimType == "ppsRap") {
				collectionName = "ppsRapDocumentReqEditors";
			} else {
				collectionName = "billingDocumentReqEditors";
			}
			
			// TODO: reuse existing editors
			if (t[collectionName]) {
				t[collectionName].destroyRecursive();
			}

			t[collectionName] = new EditorCollection({
				titleText: "Document Requirements",
				Object: DocumentReq,
				ObjectEditor: DocumentReqEditor,
				docRequirementSet: t,
				uniqueFields: ["documentType"]
			});

			if (t.object.documentReqs && t.object.documentReqs[claimType]) {
				t[collectionName].set("value", t.object.documentReqs[claimType]);
			}

			domConstruct.place(t[collectionName].domNode, attachPoint, "last");
			t[collectionName].actionStore.get("add").set("label", "Add Requirement");
			t[collectionName].startup();

			t.documentReqCollections.push(t[collectionName]);
		},

		_setActionTypeAttr: function(value) {
			/*
				value = "Edit", "Add", or "View"
			*/
			
			var t = this;

			var assignedToInsurance = false;
			if (t.object) {
				if (value == "Edit" && TableManager.get("Tables.Insurance.DocumentationRequirement.Assignment").query({ requirementSet: t.object.id }).length > 0) {
					value = "View";
					showMessage("Documentation Requirement Sets can not be modified once assigned to an insurance. Please create a copy.");
					assignedToInsurance = true;
				}
			}

			t.inherited(arguments);

			var editorCollections = t.requirementCollections.concat(t.documentReqCollections);
			arrayUtils.forEach(editorCollections, function(collection) {
				collection.set("actionType", value);
			});

			t.copyDocRequirementSetButton.set("disabled", (value != "Edit" && !assignedToInsurance));
		},

		okToSave: function() {
			var t = this;
			var ok = false;

			var nameIsUnique = true;
			if (t.actionType == "Add") {
				TableManager.get("Tables.Insurance.DocumentationRequirement.Set").query().forEach(function(set) {
					if (set.name == t.object.get("Name")) {
						nameIsUnique = false;
					}
				});
			}

			var subEditorsModified = false;
			var editorCollections = t.requirementCollections.concat(t.documentReqCollections);
			arrayUtils.forEach(editorCollections, function(collection) {
				subEditorsModified = subEditorsModified || collection.get("modified");
			});

			if (!t.formNode.validate()) {
				showMessage("Please correct any invalid fields.");
			} else if (query(".pinkBackground").length > 0) {
				showMessage("Please fill in all required fields.");
			} else if (t.actionType === "Edit" && !t.get("modified") && !subEditorsModified) {
				showMessage("You have not made any changes.");
			} else if (!nameIsUnique) {
				showMessage('"' + t.object.get("Name") + '" is already in use. Please choose a different name.');
			} else if (false) {

			} else {
				ok = true;
			}
			return ok;
		},

		save: function() {
			var t = this;
			var deferred = new Deferred();
			var promise = deferred.promise;

			if (t.okToSave()) {
				var saveData = t._getSaveData();

				var options = { saveData: saveData };

				promise = serverRequest("Tables.Insurance.DocumentationRequirement.Set", "Save", options, false, 15000).then(function() {
				}, function(error) {
					showMessage(error);
					throw error;
				});
			} else {
				deferred.reject();
			}
			return promise;
		},

		_getSaveData: function() {
			var t = this;
			var saveData = t.object.getData();

			saveData.requirementsOptions = {};
			if (t.ppsRapRequirementEditors && !t.ppsRapRequirementEditors._beingDestroyed) {
				saveData.requirementsOptions.ppsRap = t.ppsRapRequirementEditors.get("value");
			}

			saveData.requirementsOptions.billingType = {};
			if (t.billingTypeRequirementEditors && !t.billingTypeRequirementEditors._beingDestroyed) {
				saveData.requirementsOptions.billingType = t.billingTypeRequirementEditors.get("value");
			}

			saveData.documentReqs = {};
			if (t.ppsRapDocumentReqEditors && !t.ppsRapDocumentReqEditors._beingDestroyed) {
				saveData.documentReqs.ppsRap = t.ppsRapDocumentReqEditors.get("value");
			}

			saveData.documentReqs.billingType = {};
			if (t.billingDocumentReqEditors && !t.billingDocumentReqEditors._beingDestroyed) {
				saveData.documentReqs.billingType = t.billingDocumentReqEditors.get("value");
			}
			return saveData;
		}
	});

	var requirementModel = new ObjectModel({ propertyDefinitions: [
		{ name: "id", type: "text" },
		{ name: "defaultValue", type: "text" },
		{ name: "requirement", type: "enum", displayIndex: requirementOptions }
	]});

	var requirementTemplate = {
		noForm: true,
		containers: [{
			sections: [{
				hideDivider: true,
				rows: [{
					fields: [{
						name: "requirement",
						labelSize: "grander"
					}]
				}]
			}]
		}]
	};

	var Requirement = requirementModel.getObject();
	var RequirementEditor = declare([ObjectEditor], {
		objectModel: requirementModel,
		templateString: template.generate(requirementTemplate),

		postMixInProperties: function() {
			var t = this;
			t.inherited("postMixInProperties", arguments);

			t.getPropertyModel("requirement").noDefault = true;
		},

		buildRendering: function() {
			var t = this;
			t.inherited("buildRendering", arguments);

			t.require("requirement");
		},

		postCreate: function() {
			var t = this;
			t.inherited(arguments);

			t.own(aspect.before(t, "_setObjectAttr", function(object) {
				// default value
				if (!object.requirement) {
					object.requirement = object.defaultValue;
				}
			}));
		}
	});

	var documentReqModel = new ObjectModel({ propertyDefinitions: [
		{ name: "id", type: "text" },
		{ name: "documentType", type: "tableItem", tableName: "DocumentLibrary.Type", labelAttribute: "Name", required: true },
		{ name: "dates", type: "enum" }
	]});

	var documentReqTemplate = {
		noForm: true,
		containers: [{
			sections: [{
				hideDivider: true,
				rows: [{
					fields: [{
						name: "documentType",
						displayName: "Document Type:",
						labelSize: "medium-large"
					}, {
						name: "dates",
						hideLabel: true
					}]
				}]
			}]
		}]
	};

	var DocumentReq = documentReqModel.getObject();
	var DocumentReqEditor = declare([ObjectEditor], {
		objectModel: documentReqModel,
		templateString: template.generate(documentReqTemplate),
		datesButton: {},

		postMixInProperties: function() {
			var t = this;
			t.inherited("postMixInProperties", arguments);

			t.getPropertyModel("dates").fieldProperties.multiple = true;

			t.getPropertyModel("dates").getField = function(srcNodeRef) {
				return new CheckedMultiSelect(this.getFieldProperties(),srcNodeRef);
			};
		},

		buildRendering: function() {
			var t = this;
			t.inherited("buildRendering", arguments);
		},

		postCreate: function() {
			var t = this;
			t.inherited(arguments);

			t.own(aspect.after(t, "_setObjectAttr", function(object) {
				if (typeof object.dates != "undefined") {
					t.datesField.set("value", object.dates.split(","));
				}
			}, true));

			t.registerChangeLogic("documentType", function(value) {
				arrayUtils.forEach(t.datesField.getOptions(), function(option) {
					t.datesField.removeOption(option.value);
				});

				TableManager.get("DocumentLibrary.DateTypeRelation").query({ Type: value }).forEach(function(dateTypeRelation) {
					var dateId = dateTypeRelation.DateField;
					var dateName = TableManager.get("DocumentLibrary.Date").query({ Id: dateId })[0].Name;
					t.datesField.addOption(domConstruct.create("option", {label:escapeText(dateName), value:dateId}));
				});

				if (t.collection) {
					t.collection.updateUniqueFields();
				}

				var datesMenu = new TooltipDialog({
					content: t.datesField
				});

				t.datesButton.set("dropDown", datesMenu);
			});
			t.datesButton = new DropDownButton({
				dropDown: new TooltipDialog(),
				label: "Dates"
			});
			domConstruct.place(t.datesButton.domNode, t.datesFieldContainer, "last");

			var removeLink = domConstruct.create("span", {
				innerHTML: "Remove",
				style: "text-decoration: underline; outline: 0; cursor: pointer;"
			}, t.datesFieldContainer, "last");
			
			t.own(on(removeLink, "click", function() {
				if (t.actionType != "View") {
					t.collection.remove(t.id);
				}
			}));
		},

		_getValueAttr: function() {
			// summary:
			//		Support get('value')

			var t = this;
			var value = t.object.getData();

			return value;
		}
	});

	return DocRequirementSetEditor;
});
