/*
	thornberry/patientStatus/LocationOfCareWidget.js
	
	Author: *Unknown*
	Created: *Unknown*
	Purpose:
	History:
	2021-09-20 WRA ND-11514 In getCbsa(), use the most up-to-date FIPS data if there is no effectiveYear match.
	2021-03-11 WRA ND-10540 Added careSite field for homecare patients (billing component)
	2021-03-10 WRA ND-10692 Added getCbsa() to use FIPS-based CBSA codes.
*/
define([
	"dojo/_base/declare",
	"thornberry/object/ObjectEditor",
	"thornberry/object/ObjectModel",
	"thornberry/object/TableManager",
	"dojo/_base/array",
	"dojo/_base/lang",
	"dojo/aspect",
	"dojo/date",
	"dojo/date/locale",
	"dojo/dom-style",
	"dojo/dom-class",
	"dojo/on",
	"dijit/registry",
	"dijit/form/Button",
	"gridx/Grid",
	"gridx/core/model/cache/Sync",
	"gridx/modules/IndirectSelect",
	"gridx/modules/select/Row",
	"gridx/modules/RowHeader",
	"dijit/_WidgetBase",
	"dijit/_TemplatedMixin",
	"dijit/_WidgetsInTemplateMixin"
], function (declare, ObjectEditor, ObjectModel, TableManager, arrayUtils, lang, aspect, date, locale, domStyle, domClass, on, registry,
	Button, Grid, Cache, IndirectSelect, SelectRow, RowHeader, _WidgetBase, _TemplatedMixin, _WidgetsInTemplateMixin) {

	var fieldProperties = [
		{
			name: "StartDate",
			displayName: "Start Date",
			type: "date",
			required: true
		},
		{
			name: "Location",
			displayName: "Location",
			type: "tableItem",
			tableName: "Tables.Location",
			required: true
		},
		{
			name: "Program",
			displayName: "Program",
			type: "tableItem",
			tableName: "Tables.Program",
			required: true
		},
		{
			name: "Municipality",
			displayName: "Municipality",
			type: "tableItem",
			tableName: "Tables.Municipality",
			required: true
		},
		{
			name: "CBSA",
			displayName: "CBSA",
			type: "text"
		},
		{
			name: "Area",
			displayName: "Area",
			type: "text"
		},
		{
			name: "CareSite",
			displayName: "Care Site",
			type: "tableItem",
			tableName: "Tables.CareLocation"
		},
		{
			name: "PlaceOfService",
			displayName: "Place Of Service",
			type: "tableItem",
			tableName: "Tables.PlaceOfService"
		}
	];

	var fieldModel = new ObjectModel({ propertyDefinitions: fieldProperties });
	var LocationOfCare = fieldModel.getObject();

	var fieldTemplate = '<div>' +
		'<div class="content-flex-container">' +
		'<div class="flex-field-container" data-dojo-attach-point="StartDateContainer">' +
		'<div class="field-name field-name-large" data-dojo-attach-point="StartDateLabelContainer">' +
		'<label>Start Date</label>' +
		'</div>' +
		'<div class="field-value field-value-medium" data-dojo-attach-point="StartDateFieldContainer">' +
		'<input data-dojo-attach-point="StartDateNode">' +
		'</div>' +
		'</div>' +
		'</div>' +
		'<div class="content-flex-container">' +
		'<div class="flex-field-container" data-dojo-attach-point="LocationContainer">' +
		'<div class="field-name field-name-large" data-dojo-attach-point="LocationLabelContainer">' +
		'<label>Location</label>' +
		'</div>' +
		'<div class="field-value field-value-medium" data-dojo-attach-point="LocationFieldContainer">' +
		'<input data-dojo-attach-point="LocationNode">' +
		'</div>' +
		'</div>' +
		'</div>' +
		'<div class="content-flex-container">' +
		'<div class="flex-field-container" data-dojo-attach-point="ProgramContainer">' +
		'<div class="field-name field-name-large" data-dojo-attach-point="ProgramLabelContainer">' +
		'<label>Program</label>' +
		'</div>' +
		'<div class="field-value field-value-medium" data-dojo-attach-point="ProgramFieldContainer">' +
		'<input data-dojo-attach-point="ProgramNode">' +
		'</div>' +
		'</div>' +
		'</div>' +
		'<div class="content-flex-container">' +
		'<div class="flex-field-container" data-dojo-attach-point="CareSiteContainer">' +
		'<div class="field-name field-name-large" data-dojo-attach-point="CareSiteLabelContainer">' +
		'<label>Care Site</label>' +
		'</div>' +
		'<div class="field-value field-value-medium" data-dojo-attach-point="CareSiteFieldContainer">' +
		'<input data-dojo-attach-point="CareSiteNode">' +
		'</div>' +
		'</div>' +
		'</div>' +
		'<div class="content-flex-container">' +
		'<div class="flex-field-container" data-dojo-attach-point="MunicipalityContainer">' +
		'<div class="field-name field-name-large" data-dojo-attach-point="MunicipalityLabelContainer">' +
		'<label>Municipality</label>' +
		'</div>' +
		'<div class="field-value field-value-medium" data-dojo-attach-point="MunicipalityFieldContainer">' +
		'<input data-dojo-attach-point="MunicipalityNode">' +
		'</div>' +
		'</div>' +
		'</div>' +
		'<div class="content-flex-container">' +
		'<div class="flex-field-container" data-dojo-attach-point="CBSAContainer">' +
		'<div class="field-name field-name-large" data-dojo-attach-point="CBSALabelContainer">' +
		'<label>CBSA</label>' +
		'</div>' +
		'<div class="field-value field-value-small" data-dojo-attach-point="CBSAFieldContainer">' +
		'<input data-dojo-attach-point="CBSANode">' +
		'</div>' +
		'</div>' +
		'</div>' +
		'<div class="content-flex-container">' +
		'<div class="flex-field-container" data-dojo-attach-point="AreaContainer">' +
		'<div class="field-name field-name-large" data-dojo-attach-point="AreaLabelContainer">' +
		'<label>Area</label>' +
		'</div>' +
		'<div class="field-value field-value-medium" data-dojo-attach-point="AreaFieldContainer">' +
		'<input data-dojo-attach-point="AreaNode">' +
		'</div>' +
		'</div>' +
		'</div>' +
		'<div class="content-flex-container">' +
		'<div class="flex-field-container" data-dojo-attach-point="PlaceOfServiceContainer">' +
		'<div class="field-name field-name-large" data-dojo-attach-point="PlaceOfServiceLabelContainer">' +
		'<label>Place Of Service</label>' +
		'</div>' +
		'<div class="field-value field-value-medium" data-dojo-attach-point="PlaceOfServiceFieldContainer">' +
		'<input data-dojo-attach-point="PlaceOfServiceNode">' +
		'</div>' +
		'</div>' +
		'</div>' +
		'</div>';

	var LocationOfCareFields = declare([ObjectEditor], {
		objectModel: fieldModel,
		templateString: fieldTemplate,

		postMixInProperties: function () {
			var t = this;
			t.inherited("postMixInProperties", arguments);
		},

		buildRendering: function () {
			var t = this;
			t.inherited("buildRendering", arguments);
		},

		postCreate: function () {
			var t = this;
			t.inherited(arguments);

			t.own(aspect.after(t, "_setObjectAttr", function (object) {
				// Find existing Location's NPI & Branch ID to ensure the values match when changing Location
				var patientLocationOfCareStore = TableManager.get("patientLocationOfCareStore");
				var locationId = "";
				var locationOject;
				var npi = "";
				var branchId = "";
				var lastIndex;
				if (typeof object.row !== "undefined") {
					lastIndex = object.row - 2; // subtract one to turn into 0-based index; subtract another to find the line before the one being edited
				} else {
					lastIndex = patientLocationOfCareStore.data.length - 1;
				}
				if (lastIndex >= 0) {
					locationId = patientLocationOfCareStore.data[lastIndex].Location;
					if (locationId != "") {
						locationOject = TableManager.get("Tables.Location").get(locationId);
						npi = locationOject.npi;
						branchId = locationOject.branchId;
					}
				}

				var restrictIdArray = [];

				if (t.LocationField) {
					restrictIdArray = [];
					TableManager.get("Tables.Location").query((patientType === "Homecare" ? { locationType: 1 } : (patientType === "Hospice" ? { locationType: 2 } : null))).forEach(function (location) {
						if (((npi == "") || (location.npi === npi)) && ((branchId == "") || (location.branchId === branchId))) {
							restrictIdArray.push(location.id);
						}
					});
					// Include any location's previously used
					for (var i = 0; i < patientLocationOfCareStore.data.length; i++) {
						locationId = patientLocationOfCareStore.data[i].Location;
						if (arrayUtils.indexOf(restrictIdArray, locationId) === -1) {
							restrictIdArray.push(locationId);
						}
					}
					t.populateSelect(t.LocationField, TableManager.get("Tables.Location"), t.object.get("Location"), true, restrictIdArray);
				}

				if (t.HospiceLocationField) {
					restrictIdArray = [];
					TableManager.get("Tables.PatientStatusLocation").query({ HasSiteOfService: "Yes" }).forEach(function (location) {
						restrictIdArray.push(location.id);
					});
					t.populateSelect(t.HospiceLocationField, TableManager.get("Tables.PatientStatusLocation"), t.object.get("HospiceLocation"), true, restrictIdArray);
				}

				if (patientType === "Homecare") {
					t.require("CareSite");
					t.updateFieldPinking();
				} else {
					domStyle.set(t.CareSiteContainer, "display", "none");
				}
			}, true));

			if (t.HospiceLocationField && t.HospiceFacilityField) {
				t.registerChangeLogic("HospiceLocation", function (value) {
					var restrictIdArray = [];
					TableManager.get("Tables.OutsideTables.Facility").query().forEach(function (facility) {
						if (facility.PatientStatusLocation === value || facility.PatientStatusLocation === "" || facility.PatientStatusLocation === 0) {
							restrictIdArray.push(facility.id);
						}
					});
					t.populateSelect(t.HospiceFacilityField, TableManager.get("Tables.OutsideTables.Facility"), (t.inObjectLoad ? t.object.get("HospiceFacility") : ""), true, restrictIdArray);
				});
			}

			if (t.MunicipalityField && t.CBSAField) {
				var updateCbsa = function() {
					var municipality = t.object.get("Municipality");
					var startDate = t.object.get("StartDate");

					if (municipality && municipality != "") {
						t.object.set("CBSA", getCbsa(municipality, startDate));
					} else {
						t.object.set("CBSA", "");
					}
				};

				t.registerChangeLogic("Municipality", function() {
					updateCbsa();
				});

				t.registerChangeLogic("StartDate", function () {
					updateCbsa();
				});
			}

			/*if (t.CBSAField) {
				t.registerChangeLogic("CBSA", function (value) {
					debugger;
				});
			}*/
		},

		_getValueAttr: function () {
			// summary:
			//		Support get('value')

			var t = this;
			var value = {};
			var changed = true;
			var locationOfCareSaveString = "";
			var locationOfCareWidget = registry.byId(t.locationOfCareWidgetId);
			var action = locationOfCareWidget.action;

			if (t.object && action !== "") {
				value = t.object.getData();

				value.action = action;

				if (action === "Edit" || action === "Delete") {
					value.id = locationOfCareWidget.grid.select.row.getSelected()[0];
				}

				if (action === "Add" || action === "Edit") {
					// Check for changes
					if (action === "Edit") {
						var oldObj = TableManager.get("patientLocationOfCareStore").get(value.id);
						if (value.StartDate === oldObj.StartDate &&
							value.Location === oldObj.Location.toString() &&
							value.Program === oldObj.Program.toString() &&
							value.Municipality === oldObj.Municipality.toString() &&
							value.Area === oldObj.Area.toString() &&
							value.CareSite === oldObj.CareSite.toString() && 
							value.PlaceOfService === oldObj.PlaceOfService.toString()) {
							changed = false;
						}
					}
				}

				if (changed) {
					locationOfCareSaveString = JSON.stringify(value);
				}
			}

			return locationOfCareSaveString;
		},

		_setDisabledAttr: function (disable) {
			var t = this;
			if (typeof disable === "undefined") {
				disable = true;
			}

			t.disable("StartDate", disable);
			t.disable("Location", disable);
			t.disable("Program", disable);
			t.disable("Municipality", disable);
			t.disable("CBSA", true);
			t.disable("Area", disable);
			t.disable("CareSite", disable);
			t.disable("PlaceOfService", disable);
			if (disable && t.object) {
				t.object.set("StartDate", "");
				t.object.set("Location", "");
				t.object.set("Program", "");
				t.object.set("Municipality", "");
				t.object.set("CBSA", "");
				t.object.set("Area", "");
				t.object.set("CareSite", "");
				t.object.set("PlaceOfService", "");
			}

			t.updateFieldPinking();
		},

		validate: function () {
			var t = this;
			var valid = true;
			for (var property in t.objectModel.propertyDefinitions) {
				var fieldName = t.objectModel.propertyDefinitions[property].name;

				var widget = t[fieldName + "Field"];
				if (widget && typeof widget.validate !== "undefined") {
					valid = valid && widget.validate();
				}
			}
			return valid;
		}
	});

	var LocationOfCareWidget = declare([_WidgetBase, _TemplatedMixin, _WidgetsInTemplateMixin], {
		templateString: '<div class="content-area" data-dojo-attach-point="locationOfCareContainer">' +
			'<div class="content-flex-container" data-dojo-attach-point="actions">' +
			'<button data-dojo-type="dijit/form/Button" type="button" data-dojo-attach-point="btnAdd" disabled>Add</button>' +
			'<button data-dojo-type="dijit/form/Button" type="button" data-dojo-attach-point="btnEdit" disabled>Edit</button>' +
			'<button data-dojo-type="dijit/form/Button" type="button" data-dojo-attach-point="btnDelete" disabled>Delete</button>' +
			'<div data-dojo-attach-point="locDeleteMessage" class="label" style="width:540px;display:none">The selected location of care has been marked for deletion. Click "Save" to complete. Click "Cancel" to undo.</div>' +
			'</div>' +
			'<div class="content-flex-container" data-dojo-attach-point="gridContainer" style="margin-bottom: 5px;"></div>' +
			'<div class="content-flex-container" data-dojo-attach-point="fieldsContainer"></div>' +
			'</div>',
		grid: {},
		fields: null,
		action: "",

		postMixInProperties: function () {
			var t = this;
			t.inherited("postMixInProperties", arguments);
		},

		buildRendering: function () {
			var t = this;
			t.inherited("buildRendering", arguments);

			var getFromStore = function (store, id, field) {
				var val = "";
				var item = TableManager.get(store).get(id);
				if (item) {
					if (item[field]) {
						val = item[field];
					}
				}
				return escapeText(val);
			};

			var dateFormatter = function (value) {
				value = value[this.field];
				return (value ? locale.format(locale.parse(value, { selector: "date", datePattern: "yyyy-MM-dd" }), { selector: "date", datePattern: "MM/dd/yyyy" }) : "");
			};

			var layout = [];
			layout.push({ id: "row", field: "row", name: "Seq#", width: "25px" });
			layout.push({ id: "startDate", field: "StartDate", name: "Start Date", width: "80px", formatter: dateFormatter });
			layout.push({ id: "stopDate", field: "StopDate", name: "End Date", width: "80px", formatter: dateFormatter });
			layout.push({ id: "days", field: "Days", name: "Days", width: "60px" });
			layout.push({ id: "location", field: "Location", name: "Location", width: "175px", formatter: function (itemData) { return getFromStore("Tables.Location", itemData.Location, "name"); } });
			layout.push({ id: "program", field: "Program", name: "Program", width: "175px", formatter: function (itemData) { return getFromStore("Tables.Program", itemData.Program, "name"); } });
			layout.push({ id: "municipality", field: "Municipality", name: "Municipality", width: "175px", formatter: function (itemData) { return getFromStore("Tables.Municipality", itemData.Municipality, "name"); } });
			layout.push({ id: "cbsa", field: "CBSA", name: "CBSA", width: "60px", formatter: function (itemData) { return getCbsa(itemData.Municipality, locale.parse(itemData.StartDate, { selector: "date", datePattern: "yyyy-MM-dd" })) } });
			layout.push({ id: "area", field: "Area", name: "Area", width: "60px" });
			layout.push({ id: "placeOfService", field: "PlaceOfService", name: "Place Of Service", width: "175px", formatter: function (itemData) { return getFromStore("Tables.PlaceOfService", itemData.PlaceOfService, "name"); } });

			if (patientType === "Homecare") {
				layout.push({ id: "careSite", field: "CareSite", name: "Care Site", width: "175px", formatter: function (itemData) { return getFromStore("Tables.CareLocation", itemData.CareSite, "name"); } });
			}

			var patientLocationOfCareStore = TableManager.get("patientLocationOfCareStore");

			if (!(t.grid && t.grid.domNode)) {
				var grid = new Grid({
					cacheClass: Cache,
					store: patientLocationOfCareStore,
					structure: layout,
					autoHeight: true,
					selectRowMultiple: false,
					selectRowTriggerOnCell: true,
					modules: [
						SelectRow,
						RowHeader,
						IndirectSelect
					]
				});

				domStyle.set(grid.domNode, "width", "882px");
				t.gridContainer.appendChild(grid.domNode);
				grid.startup();
				t.grid = grid;
			}

			if (!t.fields) {
				t.fields = new LocationOfCareFields({ locationOfCareWidgetId: t.id });
				t.fieldsContainer.appendChild(t.fields.domNode);
			}
			t.fields.set("object", new LocationOfCare());

			t.grid.connect(t.grid.select.row, "onSelected", lang.hitch(t, t.updateActions));
			t.grid.connect(t.grid, "onCellDblClick", lang.hitch(t, t.actionEdit));

			t.own(aspect.after(registry.byId(t.statusEditorId), "_setObjectAttr", lang.hitch(t, function () {
				t.updateStartDateConstraints();

				t.grid.connect(t.grid.body, "onAfterRow", lang.hitch(t, function (row) {
					var currentSOCDate = registry.byId(t.statusEditorId).object.get("agencySocDate");
					if (currentSOCDate) {
						var item = row.grid.model.byId(row.id).item;
						var locDate = locale.parse(item.StartDate, { selector: "date", datePattern: "yyyy-MM-dd" });
						if (locDate < currentSOCDate) {
							domClass.add(row.node(), "invalidRow");
						}
					}
				}));
			})));
		},

		postCreate: function () {
			var t = this;
			t.inherited("postCreate", arguments);
			var patientLocationOfCareStore = TableManager.get("patientLocationOfCareStore");

			t.set("disabled", true);
			t.fields.set("disabled", true);

			if (registry.byId(t.statusEditorId).actionType === "View") {
				domStyle.set(t.actions, "display", "none");
				domStyle.set(t.fieldsContainer, "display", "none");
			}

			t.own(on(t.btnAdd, "click", lang.hitch(t, t.actionAdd)),
				on(t.btnEdit, "click", lang.hitch(t, t.actionEdit)),
				on(t.btnDelete, "click", lang.hitch(t, t.actionDelete)));

			t.grid.model.clearCache();
			t.grid.body.refresh();
			t.grid.resize();
		},

		_setDisabledAttr: function (disable) {
			var t = this;
			if (typeof disable === "undefined") {
				disable = true;
			}

			if (disable) {
				var patientLocationOfCareStore = TableManager.get("patientLocationOfCareStore");
				for (var i = 0; i < patientLocationOfCareStore.data.length; i++) {
					t.grid.select.row.unselectable[patientLocationOfCareStore.data[i].id] = 1;
				}
			} else {
				t.grid.select.row.unselectable = {};
			}
			t.btnAdd.set("disabled", disable);
			if (disable) {
				t.btnEdit.set("disabled", disable);
				t.btnDelete.set("disabled", disable);
			}
		},

		updateActions: function () {
			var t = this;
			var selection = t.grid.select.row.getSelected();
			if (t.action === "" && selection.length > 0) {
				t.btnEdit.set("disabled", false);
				t.btnDelete.set("disabled", false);
			} else {
				t.btnEdit.set("disabled", true);
				t.btnDelete.set("disabled", true);
			}
		},

		updateStartDateConstraints: function () {
			var t = this;
			var statusEditor = registry.byId(t.statusEditorId);
			if (statusEditor.object) {
				var dischargeDate = statusEditor.object.get("dischargeDate");
				var snaDate = statusEditor.object.get("snaDate");
				var agencySocDate = statusEditor.object.get("agencySocDate");
				var socDate = locale.parse(patientInfo.socDate, { selector: "date", datePattern: "yyyy-MM-dd" });

				var max = null;
				if (patientInfo.patientStatus === 7 && dischargeDate !== "") {
					max = dischargeDate;
				} else if (patientInfo.patientStatus === 5 && snaDate !== "") {
					max = snaDate;
				}

				var min = null;
				if (agencySocDate) {
					min = agencySocDate;
				} else if (socDate !== "") {
					min = socDate;
				}

				var startDateField = t.fields.StartDateField;

				var constraints = {};
				if (max) {
					constraints.max = max;
				}
				if (min) {
					constraints.min = min;
				}
				startDateField.set("constraints", constraints);
			}
		},

		validateLocationOfCareItems: function () {
			// TODO: This needs updated to allow for the entry of up to one Location of Care when SOC has not yet occurred (homecare patients)
			var t = this;
			var statusEditor = registry.byId(t.statusEditorId);
			if (statusEditor.object) {
				var currentSOCDate = statusEditor.object.get("agencySocDate");
				if (currentSOCDate) {
					var patientLocationOfCareStore = TableManager.get("patientLocationOfCareStore");
					for (var i = 0; i < patientLocationOfCareStore.data.length; i++) {
						var locDate = locale.parse(patientLocationOfCareStore.data[i].StartDate, { selector: "date", datePattern: "yyyy-MM-dd" });
						if (locDate < currentSOCDate) {
							actionShowMessage("This patient's levels of care are prior to this start of care. Please review and correct any discrepancies.");
							break;
						}
					}
				}

				t.grid.model.clearCache();
				t.grid.body.refresh();
			}
		},

		validate: function () {
			return this.fields.validate();
		},

		actionAdd: function () {
			var t = this;
			t.action = "Add";

			t.set("disabled", true);
			t.fields.set("disabled", false);
			t.grid.select.row.clear();

			t.fields.set("object", new LocationOfCare());
		},

		actionEdit: function () {
			var t = this;
			var selectedId = t.grid.select.row.getSelected()[0];
			if (selectedId && t.action === "") {
				t.action = "Edit";
				t.set("disabled", true);
				t.fields.set("disabled", false);

				var selectedLOC = TableManager.get("patientLocationOfCareStore").get(selectedId);
				t.fields.set("object", new LocationOfCare(selectedLOC));
			}
		},

		actionDelete: function () {
			var t = this;
			if (patientInfo.patientStatus === 6 && TableManager.get("patientLocationOfCareStore").data.length === 1) {
				actionShowMessage("You must have at least one level of care for an admitted patient.");
			} else {
				t.action = "Delete";
				t.set("disabled", true);
				domStyle.set(t.locDeleteMessage, "display", "inline");
			}
		},

		_getValueAttr: function () {
			// summary:
			//		Support get('value')

			var t = this;
			var value = "";

			if (t.fields && t.action !== "") {
				value = t.fields.get("value");
			}

			return value;
		}
	});

	var getCbsa = function (municipality, startDate) {
		var cbsa = "";

		if (startDate == null) {
			startDate = new Date();
		}

		if (patientType === "Hospice") {
			startDate = date.add(startDate, "quarter", 1);
		}

		var effectiveYear = startDate.getFullYear();

		// Use most up-to-date EffectiveYear
		var maxEffectiveYear = null;
		arrayUtils.forEach(TableManager.get("cbsaStore").query({}), function(obj) {
			if (obj.EffectiveYear <= effectiveYear || effectiveYear == null) {
				maxEffectiveYear = Math.max(obj.EffectiveYear, maxEffectiveYear);
			}
		});
		effectiveYear = maxEffectiveYear;

		var item = TableManager.get("cbsaStore").query({ Municipality: municipality, EffectiveYear: effectiveYear })[0];
		if (item) {
			cbsa = item.CBSA;
		} else {
			cbsa = TableManager.get("Tables.Municipality").get(municipality)["cbsa"];
		}

		return cbsa;
	}

	return LocationOfCareWidget;
}
);