define([
	"dojo/_base/declare",
	"dojo/_base/lang",
	"dojo/_base/array",
	"thornberry/layout/View",
	"dojo/store/Memory",
	"dojo/aspect",
	"dojo/dom",
	"dojo/dom-attr",
	"dojo/string",
	"dojo/dom-style",
	"dojo/on",
	"dojo/dom-class",
	"dojo/dom-construct",
	"dijit/form/Button",
	"thornberry/report/formatter",
	"thornberry/serverRequest",
	"thornberry/editor/load",
	"gridx/Grid",
	"gridx/core/model/cache/Sync",
	"gridx/modules/select/Row",
	"gridx/modules/extendedSelect/Row",
	"gridx/modules/IndirectSelectColumn",
	"gridx/modules/Filter",
	"gridx/modules/SingleSort",
	"gridx/core/model/extensions/FormatSort",
	"gridx/modules/VirtualVScroller",
	"gridx/modules/ColumnResizer",
	"gridx/modules/HiddenColumns",
	"gridx/modules/move/Column",
	"gridx/modules/Bar",
	"gridx/support/Summary",
	"gridx/modules/pagination/PaginationBar",
	"gridx/modules/Dod",
	"dojo/i18n!gridx/nls/SummaryBar",
	"dojo/Deferred",
	"dojo/when",
	"dojo/promise/all",
	"dojo/json",
	"dojo/query",
	"dojo/_base/html",
],function(declare,lang,arrayUtils,View,Memory,aspect,dom,domAttr,string,domStyle,on,domClass,domConstruct,Button,formatter,serverRequest,load,Grid,Cache,SelectRow,ExtendedSelectRow,IndirectSelectColumn,Filter,SingleSort,FormatSort,VirtualVScroller,ColumnResizer,HiddenColumns,MoveColumn,Bar,Summary,PaginationBar,Dod,nls,Deferred,when,all,json,query,html) {
	var DEFAULTGRIDPAGESIZE = 25;
	
	// Define global function used by the detail formatter
	showReportDetails = function(name,column,rowId,summary) {};

	var gridRowClickHandler = function(event) {
		// Summary:
		//      When connected with the grid's onRowClick event, this causes clicking the checkbox to add the row to those already selected, and clicking anywhere else
		//      to result in only the clicked row being selected
		if (event.columnIndex !== 0 && this.select.row.enabled) {
			var selected = this.select.row.getSelected();
			if (selected.length === 1 && selected[0] === event.rowId) {
				this.select.row.deselectById(event.rowId);
			} else {
				if (!event.ctrlKey) {
					this.select.row.clear();
				}
				this.select.row.selectById(event.rowId);
			}
		}
	};

	var AltPaginationBar = declare(PaginationBar,{
		required: ["bar"],
		sizes: [10,25,50,100]
	});
	var AltSingleSort = declare(SingleSort,{
		sort: function(colId,isDescending,skipUpdateBody) {
			var t = this,
			g = t.grid,
			col = g._columnsById[colId];
			if(col && (col.sortable || col.sortable === undefined)){
				if(t._sortId != colId || t._sortDescend == !isDescending){
					t._updateHeader(colId, isDescending);
				}
				t.onSort(colId,isDescending,skipUpdateBody);
			}
		},
		onSort: function() {}
	});

	Summary.prototype.refresh = function(){
		var g = this.grid,
			sr = g.select && g.select.row,
			size = g.gridView.totalRows,
			selected = sr ? sr.getSelected().length : 0,
			tpl = sr ? nls.summaryWithSelection : nls.summary;
		this.domNode.innerHTML = string.substitute(tpl, [size, selected]);
	};

	var Pagination = declare(null,{
		gridView: null,

		constructor: function(gridView) {
			var t = this;
			t.gridView = gridView;
		},
		pageCount: function() {
			var t = this;
			return Math.max(Math.ceil(t.gridView.totalRows / t.gridView.gridPageSize), 1);
		},
		currentPage: function() {
			var t = this;
			return Math.ceil(t.gridView.gridPageStart / t.gridView.gridPageSize);
		},
		pageSize: function() {
			var t = this;
			return t.gridView.gridPageSize;
		},
		setPageSize: function(size) {
			var t = this;
			var oldSize = t.gridView.gridPageSize;
			t.gridView.gridPageSize = size;
			t.onChangePageSize(size,oldSize);
			t.gotoPage(0);
			t.gridView.controller.generalUserSettings.set("grid","pageSize",size,true);
		},
		gotoPage: function(page) {
			var t = this;
			t.gridView.page = page;
			t.gridView.gridPageStart = page * t.gridView.gridPageSize;
			if (t.gridView.pageLoaded()) {
				t.gridView.loadGridPage();
				t.gridView.refreshGrid();
			} else {
				t.gridView.loadDataPage();
			}
		},
		onSwitchPage: function() {},
		onChangePageSize: function() {}
	});
	
	var GridView = declare([View],{
		dataStore: null,        // dojo/store/Memory, stores all data from the most recent web server call
		gridStore: null,        // dojo/store/Memory, stores the current display page
		totalRows: 0,           // Stores the total number of rows without regard to paging
		dataPageStart: 0,       // The beginning index of the currently loaded set of data from the server
		dataPageSize: 1000,     // The number of rows to pull from the server on each call. The default is 1,000, but the value is actually provided by the report definition

		gridPage: 0,
		gridPageStart: 0,
		gridPageSize: DEFAULTGRIDPAGESIZE,
		pagination: null,

		editorActionButtons: null,
		
		detailColumns: null,
		selectedDetailColumns: null,
		_bodyInvalidated: false,

		postMixInProperties: function() {
			var t = this;
			t.inherited(arguments);
			
			t.dataPageSize = t.controller.report.pageSize;
			t.dataStore = new Memory();
			t.gridStore = new Memory();
			t.pagination = new Pagination(t);

			t.editorActionButtons = [];

			var userPageSize = t.controller.generalUserSettings.get("grid","pageSize");
			if (userPageSize !== "") {
				t.gridPageSize = userPageSize;
			}

			showReportDetails = function(name,column,rowId,summary) {
				var foundName = false;
				if (t.controller.report.summaryColumn) {
					var row = t.dataStore.get(rowId);
					if (row) {
						name = name + " for " + row[t.controller.report.summaryColumn];
						foundName = true;
					}
				}
				if (!foundName) {
					if (summary !== "undefined" && summary !== "") {
						name = name + " for " + summary;
					}
				}

				var detailView = new View({
					controller: t.controller,
					title: name,
					gridView: t,

					buildContent: function() {
						var t = this;
						t.backButton = new Button({
							label: "Back",
							onClick: function() {
								controller.popView();
							}
						});
						t.controller.registerToggledNode(t.backButton.domNode);
						t.backButton.placeAt(t.controller.actionContainer);

						var context = t.gridView.controller.getContext("Interactive",t.gridView.controller.report);
						context.options.column = column;
						context.options.rowId = rowId;
						context.options.includeDefinition = true;

						domStyle.set("loadingOverlay","display","block");

						serverRequest(t.controller.report.className,"GenerateDetails",context,t.controller.report.async).then(function(result) {
							domStyle.set("loadingOverlay","display","none");

							var data = result.detailData;
							var columns = result.detailDefinition;
							columns = arrayUtils.filter(columns,function(column) {
								return column.isDisplayed;
							});

							var layout = t.gridView.generateGridStructure(columns);

							t.grid = new Grid({
								cacheClass: Cache,
								style: "height:100%",
								store: new Memory({data:data}),
								structure: layout,
								modules: [
									SingleSort,
									VirtualVScroller
								]
							});
							t.domNode.appendChild(t.grid.domNode);
							t.grid.startup();
							t.grid.resize();
						},function(error) {
							domStyle.set("loadingOverlay","display","none");
							t.controller.showMessage(error);
						});
					},

					onPop: function() {
						this.defer("destroyRecursive");
					},

					refresh: function() {
						var t = this;
						domStyle.set(t.backButton.domNode,"display","");
					}
				});

				t.controller.addView(detailView);
				t.controller.pushView(detailView.id);
			};
		},

		buildContent: function() {
			// Summary:
			//      This method is responsible for setting up the content pane's content 
			var t = this;
			t.defer(function() {
				if (t.controller.report.printable()) {
					t.showList.push(t.controller.printIconNode);
				}
				if (t.controller.report.filterColumns !== "") {
					t.showList.push(t.controller.filterIconNode);
				}
				if (t.controller.report.staticColumns != 1) {
					t.showList.push(t.controller.optionIconNode);
				}
				if (t.controller.report.searchable) {
					t.showList.push(t.controller.searchIconNode);
				}
			});

			t.refreshButton = new Button({
				label: "Refresh",
				style: "margin-right: 20px",
				onClick: function() {
					t.controller.modified.summarization = true;
					t.updateGridData(true);
				}
			}).placeAt(t.controller.actionContainer);
			t.controller.registerToggledNode(t.refreshButton.domNode);
			t.controller.updateDisableList.push(t.refreshButton);

			var displayedColumns = t.controller.report.getDisplayedColumns();
			
			var modules = [
				AltSingleSort,
				Bar,
				Filter,
				ColumnResizer,
				HiddenColumns,
				MoveColumn,
				AltPaginationBar
			];
			if (t.controller.report.selectableRows) {
				modules.push(SelectRow);
				modules.push(IndirectSelectColumn);
				modules.push(ExtendedSelectRow);
			}
			
			t.detailColumns = arrayUtils.filter(displayedColumns,function(column) { return typeof column.detailWidget != "undefined"; });
			if (t.detailColumns.length > 0) {
				modules.push({
					moduleClass: Dod,
					defaultShow: true,
					useAnimation: false,
					showExpando: false,
					detailProvider: lang.hitch(t,"detailProvider")
				});
			} else {
				modules.push(VirtualVScroller);
			}
			
			t.grid = new Grid({
				gridView: t,
				cacheClass: Cache,
				store: t.gridStore,
				structure: t.generateGridStructure(displayedColumns),
				pagination: t.pagination,
				modules: modules,
				modelExtensions: [
					FormatSort
				],
				barTop: [
					{content: "<span>Search results for \"<span id='searchTextHeader'></span>\"<span id='btnClearSearch'>Clear</span></span>", className: "emphasize"}
				]
			});
			
			t.domNode.appendChild(t.grid.domNode);
			t.grid.startup();
			
			t.grid.resize();

			t.grid.connect(t.grid.columnResizer,"onResize",lang.hitch(t,"saveColumnWidth"));
			if (t.grid.select) {
				t.grid.connect(t.grid,"onRowClick",gridRowClickHandler);
				t.grid.connect(t.grid.select.row,"onSelectionChange",lang.hitch(t.controller,"refresh"));
			}

			t.grid.connect(t.grid.sort,"onSort",function() {
				var currentPreset = t.controller.presetManager.currentPreset;
				currentPreset.sort = t.getSort();
				t.controller.setCurrentPresetModified();
				t.controller.setModified("sort");
			});

			t.grid.connect(t.grid.body, "onAfterRow", function(row) {
				var item = row.grid.model.byId(row.id).item;
				try {
					arrayUtils.forEach(t.controller.report.rowClassRexExps,function(regExItem) {
						if (item[regExItem.column] && regExItem.regEx.test(item[regExItem.column])) {
							domClass.add(row.node(),regExItem.rowClass);
						}
					});
				} catch (e) {console.error(e);}
			});
			domStyle.set(t.grid.bar.topNode,"display","none");
			
			if (t.detailColumns.length) {
				t.grid.dod._detailLoadComplete = function(row){
					var _row = this._row(row), g = this.grid, escapeId = g._escapeId;
					if(!this.isShown(row)){return;}
					_row.dodLoaded = true;
					
					if(_row.defaultShow){
						html.style(_row.dodNode, 'display', 'block');
						t.invalidateBody();
						//g.body.onRender();
						t.grid.dod._row(row).dodLoaded = false;
					}
					html.style(_row.dodLoadingNode, 'display', 'none');
				};
			}
		},

		postCreate: function() {
			var t = this;
			t.inherited(arguments);

			t.own(aspect.after(t.controller,"_setSearchAttr",function(searchText) {
				dom.byId("searchTextHeader").innerHTML = escapeText(searchText);
				domStyle.set(t.grid.bar.topNode,"display",(searchText !== "" ? "" : "none"));
			},true));

			t.defer(function() {
				if (dom.byId("btnClearSearch")) {
					t.own(on(dom.byId("btnClearSearch"),"click",function() {
						t.controller.onApplyFilters();
					}));
				}
			});

			t.own(on(t,"show",function() {
				if (!t.controller.closing) {
					t.updateGridData();
				}
			}));

			t.own(aspect.after(t.controller,"setModified",function() {
				if (t.controller.viewStack.selectedChildWidget.id === t.id) {
					t.updateGridData();
				}
			}));
		},

		invalidateBody: function() {
			var t = this;
			if (!t._bodyInvalidated) {
				t._bodyInvalidated = true;
				t.defer("validateBody");
			}
		},

		validateBody: function() {
			var t = this;
			t.grid.body.onRender();
			t._bodyInvalidated = false;
		},

		updateGridData: function(force) {
			// Summary:
			//      This method should be called whenever the user makes changes that affect what data displays in the grid. It gathers
			//      all the context information and passes it to the report class's implementation of Report.Definition.Abstract:GenerateData()
			//      and then passes the results to the GridView
			var t = this;
			if (t.controller.modified.any || force) {
				t.grid.vScroller.scrollToRow(0,1);

				if (t.controller.modified.options) {
					t.setColumns(t.controller.columnPane.get("value"));
				}
				if (t.controller.modified.any) {
					t.resetPagination();
				}

				var context = t.controller.getContext("Interactive",t.controller.report);
				context.pageStart = t.get("dataPageStart");
				context.pageSize = t.get("dataPageSize");

				domStyle.set("loadingOverlay","display","block");
				
				serverRequest(t.controller.report.className,"GenerateData",context,t.controller.report.async).then(function(results) {
					t.setResults(results);
					domStyle.set("loadingOverlay","display","none");
					t.controller.clearModified();
				},function(error) {
					domStyle.set("loadingOverlay","display","none");
					t.controller.showMessage(error);
				});
			}
		},

		setResults: function(results) {
			// Summary:
			//      Takes results from Report.Definition.Abstract:GenerateData() and displays them in the grid
			var t = this;
			t.totalRows = results.total;
			t.dataStore.setData(results.data);
			t.loadGridPage();
			t.refreshGrid();
		},

		setColumns: function(columns) {
			// Summary:
			//      Takes an array of the columns that the user has selected to show, rearranges the grid columns to match, and hides any not included
			var t = this;
			var grid = t.grid;
			grid.hiddenColumns.add.apply(grid.hiddenColumns,grid.columns());
			grid.hiddenColumns.remove.apply(grid.hiddenColumns,columns);

			// Clear sorting if the sort column is no longer displayed
			if (!arrayUtils.some(columns,function(column) {
				return column == t.grid.sort._sortId;
			})) {
				t.grid.sort._sortId = "";
			}

			var columnPosition = 1;
			arrayUtils.forEach(columns,function(id) {
				var column = grid.column(id);
				if (column) {
					grid.move.column.move(column.index(),columnPosition);
					columnPosition++;
				}
			});
			
			if (t.detailColumns.length) {
				t.selectedDetailColumns = arrayUtils.filter(t.detailColumns,function(detColumn) {
					return arrayUtils.indexOf(t.controller.columnPane.get("value"),detColumn.name) !== -1;
				});
				t.grid.dod.defaultShow = t.selectedDetailColumns.length !== 0;
			}
		},

		saveColumnWidth: function(column,width) {
			// Summary:
			//      Saves a column width in the user's settings on the server
			this.controller.userSettings.set("columnWidth",column,width,true);
		},

		pageLoaded: function() {
			// Summary:
			//      Checks whether the currently loaded page of data from the server contains the display page selected by the user
			var t = this;
			var loaded = false;
			var dataStart = t.dataPageStart;
			var dataEnd = t.dataPageStart + t.get("dataPageSize");
			var gridStart = t.gridPageStart;
			var gridEnd = t.gridPageStart + t.gridPageSize;
			if ((gridStart >= dataStart) && (gridEnd <= dataEnd)) {
				loaded = true;
			}
			return loaded;
		},

		loadGridPage: function() {
			var t = this;
			var gridData = [];
			var startIndex = t.gridPageStart - t.dataPageStart;
			var endIndex = startIndex + t.gridPageSize;
			endIndex = (endIndex > t.dataStore.data.length ? t.dataStore.data.length : endIndex);
			for (var i = startIndex; i < endIndex; i++) {
				gridData.push(t.dataStore.data[i]);
			}
			t.gridStore.setData(gridData);
		},

		loadDataPage: function() {
			var t = this;
			t.dataPageStart = Math.floor(t.gridPageStart / t.get("dataPageSize")) * t.get("dataPageSize");
			t.updateGridData(true);
		},

		refreshGrid: function() {
			// Summary:
			//      After the store's data has been changed, this synchronizes the grid
			var t = this;

			if (t.grid.select && t.grid.select.row) {
				t.grid.select.row.clear();
			}
			t.grid.vScroller.scrollToRow(0,1);
			t.grid.model.clearCache();
			t.grid.body.refresh(); 

			query(".gridxDodLoadNode",t.grid.domNode).forEach(function(node) {
				domStyle.set(node,"display","none");
			});
			t.grid.model.onSizeChange();
		},

		resetPagination: function() {
			var t = this;
			t.gridPage = 0;
			t.gridPageStart = 0;
			t.dataPageStart = 0;
		},

		generateGridStructure: function(columns) {
			// Summary:
			//      Takes a thornberry/report/Report and generates the column structure from it
			var t = this;
			var structure = [];
			arrayUtils.forEach(columns,function(columnDefinition) {
				if (columnDefinition.detailWidget) {
					return;
				}
				
				// Set basic column data
				var displayName = escapeText(columnDefinition.displayName);
				/*if (columnDefinition.help !== "") {
					// TODO: handle quotes
					displayName = "<span title='"+escapeText(columnDefinition.help)+"'>"+displayName+"</span>";
				}*/

				var gridColumn = {
					id: columnDefinition.name,
					name: displayName,
					field: columnDefinition.name
				};

				// Get the column formatter
				gridColumn.formatter = formatter.getForColumn(gridColumn,columnDefinition);

				// Set the column's default width
				var width = t.controller.userSettings.get("columnWidth",columnDefinition.name);
				width = (width !== "" ? width : columnDefinition.defaultWidth || "");
				if (width !== "") {
					gridColumn.width = width + "px";
				}

				structure.push(gridColumn);
			});

			return structure;
		},

		getSort: function() {
			var t = this;
			return {column:t.grid.sort._sortId,direction:(t.grid.sort._sortDescend ? "DESC" : "ASC")};
		},

		getSelected: function() {
			var t = this;
			var selectedIds = [];
			var rows = t.grid.rows();
			for (var i = 0; i < rows.length; i++) {
				if (t.grid.select.row.isSelected(rows[i].id)) {
					selectedIds.push(rows[i].id);
				}
			}
			return selectedIds;
		},

		_getDataPageSizeAttr: function() {
			var t = this;
			return (t.dataPageSize > t.gridPageSize ? t.dataPageSize : t.gridPageSize);
		},
		
		detailProvider: function(grid, rowId, detailNode, rendered) {
			var t = this;
			if (!t.selectedDetailColumns || t.selectedDetailColumns.length === 0) {
				domStyle.set(detailNode,"border-bottom","none");
				rendered.callback();
			} else {	
				domStyle.set(detailNode,"border-bottom","1px solid #E5DAC8");	
				var rootNode = domConstruct.toDom("<div></div>");
				
				if (NOT_IE) {
					detailNode.innerHTML = ""
				}
				
				detailNode.appendChild(rootNode);		
					
				var item = t.grid.store.get(rowId);
				var promises = [];
				arrayUtils.forEach(t.selectedDetailColumns,function(column) {
					var deferred = new Deferred();
					promises.push(deferred.promise);
					require([column.detailWidget],function(Widget) {
						var jsonData = item[column.name];
						if (typeof jsonData !== "undefined" && jsonData != "") {
							try {
								var widget = new Widget({
									data: json.parse(jsonData)
								});
								rootNode.appendChild(widget.domNode);
							} catch (e) {}
							deferred.resolve();
							var row = t.grid.row(rowId,true);
							t.grid.dod._detailLoadComplete(row);   
						}
					});
					
				});
				all(promises).then(function() {
					setTimeout(function() {
						rendered.callback();
					},0);
				});
			}
		},

		refresh: function() {
			var t = this;

			domStyle.set(t.refreshButton.domNode,"display","");

			if (t.editorActionButtons.length > 0) {
				arrayUtils.forEach(t.editorActionButtons,function(button) {button.destroyRecursive();});
				t.editorActionButtons = [];
			}

			if (t.grid.select && t.grid.select.row) {
				// launch editor if enabled
				var selectedIds = t.grid.select.row.getSelected();
				if (selectedIds.length > 0 && t.controller.report.editorClassColumn !== "" && t.controller.report.layoutClassColumn !== "" && t.controller.report.editIdColumn !== "") {
					var editorClass = null;
					var layoutClass = null;
					var itemIds = [];
					var items = [];
					// Check that all items are for the same editor and layout classes if multiple
					arrayUtils.forEach(selectedIds,function(id) {
						var item = t.grid.store.get(id);
						items.push(item);
						var itemId = item[t.controller.report.editIdColumn];
						if (itemId) {
							itemIds.push(itemId);
						}
						var itemEditorClass = item[t.controller.report.editorClassColumn];
						var itemLayoutClass = item[t.controller.report.layoutClassColumn];
						editorClass = (editorClass === null || editorClass === itemEditorClass ? itemEditorClass : false);
						layoutClass = (layoutClass === null || layoutClass === itemLayoutClass ? itemLayoutClass : false);
					});
					if (itemIds.length > 0 && editorClass && layoutClass) {
						load(editorClass,layoutClass,true).then(function(editor) {
							// delete buttons again in case created in the gap when the editor was loading
							if (t.editorActionButtons.length > 0) {
								arrayUtils.forEach(t.editorActionButtons,function(button) {button.destroyRecursive();});
								t.editorActionButtons = [];
							}

							var extra = {
								reportClass: t.controller.report.className,
								itemIds: itemIds,
								items: items
							};
							var actions = editor.getObjectActions("",extra);
							arrayUtils.forEach(actions,function(action) {
								var button = new Button({
									label: action,
									onClick:function() {
										var openOrPromise = editor.onAction(action,(itemIds.length === 1 ? itemIds[0] : null),itemIds);
										when (openOrPromise,function(open) {
											if (open) {
												if (t._saveHandle) {
													t._saveHandle.remove();
												}

												t.controller.loadEditor(editor);
												t._saveHandle = on(t.controller.editorView,"save",t.refreshButton.onClick);
											} else {
												t.refreshButton.onClick();
											}
										});
									}
								});
								t.controller.registerToggledNode(button.domNode);
								domStyle.set(button.domNode,"display","");
								t.controller.actionContainer.appendChild(button.domNode);
								t.editorActionButtons.push(button);
							});
						});
					}
				}
			}
		}
	});

	return GridView;
});